----------------------------------------------------------------
--  Title     : psg_wave.vhd
--  Function  : Programmable Sound Generator (AY-3-8910/YM2149)
--  Date      : 28th,August,2000
--  Revision  : 1.01
--  Author    : Kazuhiro TSUJIKAWA (ESE Artists' factory)
----------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity psg_wave is
  port(
    pSltClk_n   : IN std_logic;
    pSltRst_n   : IN std_logic;
    PsgRegPtr   : IN std_logic_vector(3 downto 0);
    pSltDat     : INOUT std_logic_vector(7 downto 0);
    PsgAmp      : OUT std_logic_vector(7 downto 0);

    PsgRegWe    : IN std_logic
  );
end psg_wave;

architecture RTL of psg_wave is

  signal PsgClkEna   : std_logic_vector(4 downto 0);

  signal PsgFreqChA  : std_logic_vector(11 downto 0);
  signal PsgFreqChB  : std_logic_vector(11 downto 0);
  signal PsgFreqChC  : std_logic_vector(11 downto 0);
  signal PsgFreqNoise: std_logic_vector(4 downto 0);
  signal PsgChanSel  : std_logic_vector(5 downto 0);
  signal PsgVolChA   : std_logic_vector(4 downto 0);
  signal PsgVolChB   : std_logic_vector(4 downto 0);
  signal PsgVolChC   : std_logic_vector(4 downto 0);
  signal PsgFreqEnv  : std_logic_vector(15 downto 0);
  signal PsgShapeEnv : std_logic_vector(3 downto 0);

  signal PsgEdgeChA  : std_logic;
  signal PsgEdgeChB  : std_logic;
  signal PsgEdgeChC  : std_logic;
  signal PsgNoise    : std_logic;
  signal PsgVolEnv   : std_logic_vector(3 downto 0);
  signal PsgEnvReq   : std_logic;
  signal PsgEnvAck   : std_logic;

begin

  ----------------------------------------------------------------
  -- Miscellaneous control / clock enable
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

  begin

    if (pSltRst_n = '0') then

      PsgClkEna <= (others => '0');

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- Clock Enable (clock divider)
      PsgClkEna <= PsgClkEna - 1;

    end if;

  end process;

  ----------------------------------------------------------------
  -- PSG register read
  ----------------------------------------------------------------
  pSltDat <= (others => 'Z');

  ----------------------------------------------------------------
  -- PSG register write
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

  begin

    if (pSltRst_n = '0') then

      PsgFreqChA   <= (others => '1');
      PsgFreqChB   <= (others => '1');
      PsgFreqChC   <= (others => '1');
      PsgFreqNoise <= (others => '1');
      PsgChanSel   <= (others => '1');
      PsgVolChA    <= (others => '1');
      PsgVolChB    <= (others => '1');
      PsgVolChC    <= (others => '1');
      PsgFreqEnv   <= (others => '1');
      PsgShapeEnv  <= (others => '1');

      PsgEnvReq    <= '0';

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      if (PsgRegWe = '1') then
        case PsgRegPtr is
          when "0000" => PsgFreqChA(7 downto 0)  <= pSltDat;
          when "0001" => PsgFreqChA(11 downto 8) <= pSltDat(3 downto 0);
          when "0010" => PsgFreqChB(7 downto 0)  <= pSltDat;
          when "0011" => PsgFreqChB(11 downto 8) <= pSltDat(3 downto 0);
          when "0100" => PsgFreqChC(7 downto 0)  <= pSltDat;
          when "0101" => PsgFreqChC(11 downto 8) <= pSltDat(3 downto 0);
          when "0110" => PsgFreqNoise            <= pSltDat(4 downto 0);
          when "0111" => PsgChanSel              <= pSltDat(5 downto 0);
          when "1000" => PsgVolChA               <= pSltDat(4 downto 0);
          when "1001" => PsgVolChB               <= pSltDat(4 downto 0);
          when "1010" => PsgVolChC               <= pSltDat(4 downto 0);
          when "1011" => PsgFreqEnv(7 downto 0)  <= pSltDat;
          when "1100" => PsgFreqEnv(15 downto 8) <= pSltDat;
          when "1101" => PsgShapeEnv             <= pSltDat(3 downto 0);
                         PsgEnvReq               <= not PsgEnvAck;
          when others => null;
        end case;
      end if;

    end if;

  end process;

  ----------------------------------------------------------------
  -- Tone generator
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

    variable PsgCntChA : std_logic_vector(11 downto 0);
    variable PsgCntChB : std_logic_vector(11 downto 0);
    variable PsgCntChC : std_logic_vector(11 downto 0);

  begin

    if (pSltRst_n = '0') then

      PsgEdgeChA <= '0';
      PsgCntChA  := (others => '0');
      PsgEdgeChB <= '0';
      PsgCntChB  := (others => '0');
      PsgEdgeChC <= '0';
      PsgCntChC  := (others => '0');

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- Base frequency : 112kHz = 3.58MHz / 16 / 2
      if (PsgClkEna(3 downto 0) = "0000") then

        if (PsgCntChA = "000000000000") then
          PsgEdgeChA <= not PsgEdgeChA;
          if (PsgFreqChA = "000000000000") then
            PsgCntChA := "000000000000";
          else
            PsgCntChA := PsgFreqChA - 1;
          end if;
        else
          PsgCntChA := PsgCntChA - 1;
        end if;

        if (PsgCntChB = "000000000000") then
          PsgEdgeChB <= not PsgEdgeChB;
          if (PsgFreqChB = "000000000000") then
            PsgCntChB := "000000000000";
          else
            PsgCntChB := PsgFreqChB - 1;
          end if;
        else
          PsgCntChB := PsgCntChB - 1;
        end if;

        if (PsgCntChC = "000000000000") then
          PsgEdgeChC <= not PsgEdgeChC;
          if (PsgFreqChC = "000000000000") then
            PsgCntChC := "000000000000";
          else
            PsgCntChC := PsgFreqChC - 1;
          end if;
        else
          PsgCntChC := PsgCntChC - 1;
        end if;

      end if;

    end if;

  end process;

  ----------------------------------------------------------------
  -- Noise generator 
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

    variable PsgCntNoise : std_logic_vector(4 downto 0);
    variable PsgGenNoise : std_logic_vector(17 downto 0);

  begin

    if (pSltRst_n = '0') then

      PsgCntNoise := (others => '0');
      PsgGenNoise := (others => '1');
      PsgNoise    <= '1';

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- Base frequency : 112kHz = 3.58MHz / 16 / 2
      if (PsgClkEna(4 downto 0) = "00000") then

        -- Noise frequency counter
        if (PsgCntNoise /= "00000") then
          PsgCntNoise := PsgCntNoise - 1;
        elsif (PsgFreqNoise /= "00000") then
          PsgCntNoise := PsgFreqNoise - 1;
        end if;

        -- (Maximum-length linear shift resister sequence)
        -- f(x) = x^17 + x^14 + 1
        if (PsgCntNoise = "00000") then

          for I in 17 downto 1 loop
            PsgGenNoise(I) := PsgGenNoise(I - 1);
          end loop;

          if (PsgGenNoise = "00000000000000000") then
            PsgGenNoise(0) := PsgGenNoise(17) xor PsgGenNoise(14) xor '1';  -- Error trap
          else
            PsgGenNoise(0) := PsgGenNoise(17) xor PsgGenNoise(14);          -- Normal work
          end if;

          PsgNoise <= PsgGenNoise(17);

        end if;

      end if;

    end if;

  end process;

  ----------------------------------------------------------------
  -- Envelope generator
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

    variable PsgCntEnv : std_logic_vector(15 downto 0);
    variable PsgPtrEnv : std_logic_vector(4 downto 0);

  begin

    if (pSltRst_n = '0') then

      PsgEnvAck <= '0';
      PsgVolEnv <= (others => '0');

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- Base frequency : 56kHz = 3.58MHz / 8 / 2
      if (PsgClkEna(3 downto 0) = "0000") then

        if ((PsgCntEnv = "0000000000000000") or (PsgEnvAck /= PsgEnvReq)) then
          if (PsgFreqEnv = "0000000000000000") then
            PsgCntEnv := "0000000000000000";
          else
            PsgCntEnv := PsgFreqEnv - 1;
          end if;
        else
          PsgCntEnv := PsgCntEnv - 1;
        end if;

        if (PsgEnvAck /= PsgEnvReq) then
          PsgPtrEnv := "00001";
        elsif (PsgCntEnv = "0000000000000000") then
          if ((PsgPtrEnv(4) = '0') or (PsgShapeEnv(3) = '1' and PsgShapeEnv(0) = '0')) then
            PsgPtrEnv := PsgPtrEnv + 1;
          end if;
        end if;

        if (PsgEnvAck /= PsgEnvReq) then
          if (PsgShapeEnv(2) = '1') then
            PsgVolEnv <= "0000";
          else
            PsgVolEnv <= "1111";
          end if;
        elsif (PsgCntEnv = "0000000000000000") then
          case PsgPtrEnv is
            when "00000" =>
              if (PsgShapeEnv(2) = '1') then
                PsgVolEnv <= "0000";
              else
                PsgVolEnv <= "1111";
              end if;
            when "10000" =>
              if (PsgShapeEnv(3) = '0' or (PsgShapeEnv(2) xor PsgShapeEnv(1) xor PsgShapeEnv(0)) = '1') then
                PsgVolEnv <= "0000";
              else
                PsgVolEnv <= "1111";
              end if;
            when others  =>
              if ((PsgShapeEnv(2) xor (PsgShapeEnv(1) and PsgPtrEnv(4))) = '1') then
                PsgVolEnv <= PsgVolEnv + 1;
              else
                PsgVolEnv <= PsgVolEnv - 1;
              end if;
          end case;
        end if;

        PsgEnvAck <= PsgEnvReq;

      end if;

    end if;

  end process;

  ----------------------------------------------------------------
  -- Mixer control
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

    variable PsgEnaNoise : std_logic;
    variable PsgEnaTone  : std_logic;
    variable PsgEdge     : std_logic;
    variable PsgVol      : std_logic_vector(4 downto 0);
    variable PsgIndex    : std_logic_vector(3 downto 0);
    variable PsgTable    : std_logic_vector(7 downto 0);
    variable PsgMix      : std_logic_vector(9 downto 0);

  begin

    if (pSltRst_n = '0') then

      PsgMix := (others => '0');
      PsgAmp <= (others => '0');

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      case PsgClkEna(1 downto 0) is
        when "11"   =>
          PsgEnaTone  := PsgChanSel(0); PsgEdge  := PsgEdgeChA;
          PsgEnaNoise := PsgChanSel(3); PsgVol   := PsgVolChA;
        when "10"   =>
          PsgEnaTone  := PsgChanSel(1); PsgEdge  := PsgEdgeChB;
          PsgEnaNoise := PsgChanSel(4); PsgVol   := PsgVolChB;
        when "01"   =>
          PsgEnaTone  := PsgChanSel(2); PsgEdge  := PsgEdgeChC;
          PsgEnaNoise := PsgChanSel(5); PsgVol   := PsgVolChC;
        when others =>
          PsgEnaTone  := '1';           PsgEdge  := '1';
          PsgEnaNoise := '1';           PsgVol   := "00000";
      end case;

      if (((PsgEnaTone or PsgEdge) and (PsgEnaNoise or PsgNoise)) = '0') then
        if (PsgVol(4) = '0') then
          PsgIndex := PsgVol(3 downto 0);
        else
          PsgIndex := PsgVolEnv;
        end if;
      else
        PsgIndex := (others => '0');
      end if;

      case PsgIndex is
        when "1111" => PsgTable := "11111111";
        when "1110" => PsgTable := "10110100";
        when "1101" => PsgTable := "01111111";
        when "1100" => PsgTable := "01011010";
        when "1011" => PsgTable := "00111111";
        when "1010" => PsgTable := "00101101";
        when "1001" => PsgTable := "00011111";
        when "1000" => PsgTable := "00010110";
        when "0111" => PsgTable := "00001111";
        when "0110" => PsgTable := "00001011";
        when "0101" => PsgTable := "00000111";
        when "0100" => PsgTable := "00000101";
        when "0011" => PsgTable := "00000011";
        when "0010" => PsgTable := "00000010";
        when "0001" => PsgTable := "00000001";
        when others => PsgTable := "00000000";
      end case;

      case PsgClkEna(1 downto 0) is
        when "00"   => PsgAmp <= PsgMix(9 downto 2);
                       PsgMix := (others => '0');
        when others => PsgMix := "00" & PsgTable + PsgMix;
      end case;

    end if;

  end process;

end RTL;
