----------------------------------------------------------------
--  Title     : psgsccr.vhd
--  Function  : PSG + SCC + RAM(EseSCC/SNATCHER compatible)
--  Date      : 28th,August,2000
--  Revision  : 1.01
--  Author    : Kazuhiro TSUJIKAWA (ESE Artists' factory)
----------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity psgsccr is
  port(
    pSltClk     : IN std_logic;
    pSltRst_n   : IN std_logic;
    pSltSltsl_n : IN std_logic;
    pSltIorq_n  : IN std_logic;
    pSltRd_n    : IN std_logic;
    pSltWr_n    : IN std_logic;
    pSltAdr     : IN std_logic_vector(15 downto 0);
    pSltDat     : INOUT std_logic_vector(7 downto 0);
    pSltBdir_n  : OUT std_logic;

    pSltCs1     : IN std_logic;
    pSltCs2     : IN std_logic;
    pSltCs12    : IN std_logic;
    pSltRfsh_n  : IN std_logic;
    pSltWait_n  : IN std_logic;
    pSltInt_n   : IN std_logic;
    pSltM1_n    : IN std_logic;
    pSltMerq_n  : IN std_logic;

    pSltClk2    : IN std_logic;
    pSltRsv5    : OUT std_logic;
    pSltRsv16   : OUT std_logic;

    pRamCeX_n   : OUT std_logic;
    pRamOeX_n   : OUT std_logic;
    pRamWeX_n   : OUT std_logic;
    pRamAdrX    : OUT std_logic_vector(18 downto 0);
    pRamDatX    : INOUT std_logic_vector(7 downto 0);

    pRamCeY_n   : OUT std_logic;
    pRamOeY_n   : OUT std_logic;
    pRamWeY_n   : OUT std_logic;
    pRamAdrY    : OUT std_logic_vector(18 downto 0);
    pRamDatY    : INOUT std_logic_vector(7 downto 0);

    pSltSndL    : OUT std_logic;
    pSltSndR    : OUT std_logic;
    pSltSound   : OUT std_logic
 );
end psgsccr;

architecture RTL of psgsccr is

  component psg_wave
    port(
      pSltClk_n : IN std_logic;
      pSltRst_n : IN std_logic;
      PsgRegPtr : IN std_logic_vector(3 downto 0);
      pSltDat   : INOUT std_logic_vector(7 downto 0);
      PsgAmp    : OUT std_logic_vector(7 downto 0);

      PsgRegWe  : IN std_logic
    );
  end component;

  component scc_wave
    port(
      pSltClk_n : IN std_logic;
      pSltRst_n : IN std_logic;
      pSltAdr   : IN std_logic_vector(7 downto 0);
      pSltDat   : INOUT std_logic_vector(7 downto 0);
      SccAmp    : OUT std_logic_vector(7 downto 0);

      SccRegWe  : IN std_logic;
      SccModWe  : IN std_logic;
      SccWavCe  : IN std_logic;
      SccWavOe  : IN std_logic;
      SccWavWe  : IN std_logic;
      SccWavWx  : IN std_logic;
      SccWavAdr : IN std_logic_vector(4 downto 0);
      SccWavDat : IN std_logic_vector(7 downto 0)
    );
  end component;

  signal pSltClk_n   : std_logic;
  signal DevHit      : std_logic;
  signal Dec1FFE     : std_logic;
  signal DecSccA     : std_logic;
  signal DecSccB     : std_logic;

  signal SccBank0    : std_logic_vector(7 downto 0);
  signal SccBank1    : std_logic_vector(7 downto 0);
  signal SccBank2    : std_logic_vector(7 downto 0);
  signal SccBank3    : std_logic_vector(7 downto 0);
  signal SccModeA    : std_logic_vector(7 downto 0);
  signal SccModeB    : std_logic_vector(7 downto 0);

  signal PsgRegPtr   : std_logic_vector(3 downto 0);
  signal PsgRegWe    : std_logic;
  signal KeyClick    : std_logic;

  signal SccRegWe    : std_logic;
  signal LSccRegWe   : std_logic;
  signal RSccRegWe   : std_logic;

  signal SccModWe    : std_logic;
  signal LSccModWe   : std_logic;
  signal RSccModWe   : std_logic;

  signal SccWavCe    : std_logic;
  signal LSccWavCe   : std_logic;
  signal RSccWavCe   : std_logic;

  signal SccWavOe    : std_logic;
  signal LSccWavOe   : std_logic;
  signal RSccWavOe   : std_logic;

  signal SccWavWe    : std_logic;
  signal LSccWavWe   : std_logic;
  signal RSccWavWe   : std_logic;

  signal SccWavWx    : std_logic;
  signal LSccWavWx   : std_logic;
  signal RSccWavWx   : std_logic;
  signal SccWavAdr   : std_logic_vector(4 downto 0);
  signal SccWavDat   : std_logic_vector(7 downto 0);

  signal SccMemCs    : std_logic;
  signal RamAdrEx    : std_logic;
  signal SccSpcCs    : std_logic;
  signal SccDmack    : std_logic;

  signal PsgAmp      : std_logic_vector(7 downto 0);
  signal SccAmp      : std_logic_vector(7 downto 0);
  signal LSccAmp     : std_logic_vector(7 downto 0);
  signal RSccAmp     : std_logic_vector(7 downto 0);

begin

  ----------------------------------------------------------------
  -- Dummy pin
  ----------------------------------------------------------------
  pSltRsv5  <= '1';
  pSltRsv16 <= '1';

  pSltClk_n <= not pSltClk;

  pSltBdir_n <= '0' when pSltSltsl_n = '0' and pSltRd_n = '0' else '1';

  ----------------------------------------------------------------
  -- Slot access control
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n, pSltIorq_n, pSltSltsl_n, pSltRd_n, pSltWr_n)

    variable DevAcs0 : std_logic;
    variable DevAcs1 : std_logic;

  begin

    if ((pSltIorq_n = '0' or pSltSltsl_n = '0') and (pSltRd_n = '0' or pSltWr_n = '0')) then
      DevAcs0 := '1';
    else
      DevAcs0 := '0';
    end if;

    if (DevAcs0 = '1' and DevAcs1 = '0') then
      DevHit <= '1';
    else
      DevHit <= '0';
    end if;

    if (pSltRst_n = '0') then
      DevAcs1 := '0';
    elsif (pSltClk_n'event and pSltClk_n = '1') then
      DevAcs1 := DevAcs0;
    end if;

  end process;

  Dec1FFE <= '1' when pSltAdr(12 downto 1) = "111111111111" else '0';
  DecSccA <= '1' when pSltAdr(15 downto 11) = "10011" and SccModeB(5) = '0' and SccBank2(5 downto 0) = "111111"
                 else '0';
  DecSccB <= '1' when pSltAdr(15 downto 11) = "10111" and SccModeB(5) = '1' and SccBank3(7) = '1'
                 else '0';

  ----------------------------------------------------------------
  -- PSG register / 1 bit sound port access
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

  begin

    if (pSltRst_n = '0') then

      PsgRegPtr <= (others => '1');
      KeyClick  <= '1';

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- I/O port access on A0h ... Resister number setting
      if (DevHit = '1' and pSltIorq_n = '0' and pSltWr_n = '0' and pSltAdr(7 downto 0) = "10100000") then
        PsgRegPtr <= pSltDat(3 downto 0);
      end if;

      -- I/O port access on AAh ... 1 bit sound port write (not PSG)
      if (DevHit = '1' and pSltIorq_n = '0' and pSltWr_n = '0' and pSltAdr(7 downto 1) = "1010101") then
        if (pSltAdr(0) = '0') then
          KeyClick <= pSltDat(7);
        elsif (pSltDat(7 downto 1) = "0000111") then
          KeyClick <= pSltDat(0);
        end if;
      end if;

    end if;

  end process;

  -- I/O port access on A1h ... Resister write
  PsgRegWe <= '1' when DevHit = '1' and pSltIorq_n = '0' and pSltWr_n = '0' and pSltAdr(7 downto 0) = "10100001" else '0';

  ----------------------------------------------------------------
  -- SCC register / wave memory access
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

  begin

    if (pSltRst_n = '0') then

      SccBank0   <= "00000000";
      SccBank1   <= "00000001";
      SccBank2   <= "00000010";
      SccBank3   <= "00000011";
      SccModeA   <= (others => '0');
      SccModeB   <= (others => '0');

      SccWavWx   <= '0';
      LSccWavWx  <= '0';
      RSccWavWx  <= '0';
      SccWavAdr  <= (others => '0');
      SccWavDat  <= (others => '0');

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      -- Mapped I/O port access on 5000-57FFh ... Bank resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 11) = "01010" and
          SccModeA(6) = '0' and SccModeA(4) = '0' and SccModeB(4) = '0') then
        SccBank0 <= pSltDat;
      end if;
      -- Mapped I/O port access on 7000-77FFh ... Bank resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 11) = "01110" and
          SccModeA(6) = '0' and SccModeA(4) = '0' and SccModeB(4) = '0') then
        SccBank1 <= pSltDat;
      end if;
      -- Mapped I/O port access on 9000-97FFh ... Bank resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 11) = "10010" and
          SccModeB(4) = '0') then
        SccBank2 <= pSltDat;
      end if;
      -- Mapped I/O port access on B000-B7FFh ... Bank resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 11) = "10110" and
          SccModeA(6) = '0' and SccModeA(4) = '0' and SccModeB(4) = '0') then
        SccBank3 <= pSltDat;
      end if;

      -- Mapped I/O port access on 7FFE-7FFFh ... Resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 13) = "011" and Dec1FFE = '1' and
          SccModeB(5 downto 4) = "00") then
        SccModeA <= pSltDat;
      end if;

      -- Mapped I/O port access on BFFE-BFFFh ... Resister write
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(15 downto 13) = "101" and Dec1FFE = '1' and
          SccModeA(6) = '0' and SccModeA(4) = '0') then
        SccModeB <= pSltDat;
      end if;

      -- Mapped I/O port access on 9860-987Fh ... Wave memory copy
      if (pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(7 downto 5) = "011" and
          DevHit = '1' and SccModeB(4) = '0' and DecSccA = '1') then
        SccWavAdr <= pSltAdr(4 downto 0);
        SccWavDat <= pSltDat;
        SccWavWx  <= '1';
        if (pSltAdr(9 downto 8) /= "11") then LSccWavWx <= '1'; end if;
        if (pSltAdr(9 downto 8) /= "10") then RSccWavWx <= '1'; end if;
      else
        SccWavWx  <= '0';
        LSccWavWx <= '0';
        RSccWavWx <= '0';
      end if;

    end if;

  end process;

  -- Mapped I/O port access on 9800-987Fh / B800-B89Fh ... Wave memory
  SccWavCe <= '1' when pSltSltsl_n = '0' and DevHit = '1' and SccModeB(4) = '0' and
                       (DecSccA = '1' or DecSccB = '1')
                  else '0';

  LSccWavCe <= '1' when SccWavCe = '1' else '0';
  RSccWavCe <= '1' when SccWavCe = '1' else '0';

  -- Mapped I/O port access on 9800-987Fh / B800-B89Fh ... Wave memory
  SccWavOe <= '1' when pSltSltsl_n = '0' and pSltRd_n = '0' and SccModeB(4) = '0' and
                       ((DecSccA = '1' and pSltAdr(7) = '0') or
                        (DecSccB = '1' and (pSltAdr(7) = '0' or pSltAdr(6 downto 5) = "00")))
                  else '0';

  LSccWavOe <= '1' when SccWavOe = '1' and pSltAdr(8) = '0' else '0';
  RSccWavOe <= '1' when SccWavOe = '1' and pSltAdr(8) = '1' else '0';

  -- Mapped I/O port access on 9800-987Fh / B800-B89Fh ... Wave memory
  SccWavWe <= '1' when pSltSltsl_n = '0' and pSltWr_n = '0' and DevHit = '1' and SccModeB(4) = '0' and
                       ((DecSccA = '1' and pSltAdr(7) = '0') or DecSccB = '1')
                  else '0';

  LSccWavWe <= '1' when SccWavWe = '1' and pSltAdr(9 downto 8) /= "11" else '0';
  RSccWavWe <= '1' when SccWavWe = '1' and pSltAdr(9 downto 8) /= "10" else '0';

  -- Mapped I/O port access on 9880-988Fh / B8A0-B8AF ... Resister write
  SccRegWe <= '1' when pSltSltsl_n = '0' and pSltWr_n = '0' and
                       ((DecSccA = '1' and pSltAdr(7 downto 5) = "100") or
                        (DecSccB = '1' and pSltAdr(7 downto 5) = "101")) and
                       DevHit = '1' and SccModeB(4) = '0'
                  else '0';

  LSccRegWe <= '1' when SccRegWe = '1' and pSltAdr(9 downto 8) /= "11" else '0';
  RSccRegWe <= '1' when SccRegWe = '1' and pSltAdr(9 downto 8) /= "10" else '0';

  -- Mapped I/O port access on 98C0-98FFh / B8C0-B8DFh ... Resister write
  SccModWe <= '1' when pSltSltsl_n = '0' and pSltWr_n = '0' and pSltAdr(7 downto 6) = "11" and
                       (DecSccA = '1' or (pSltAdr(5) = '0' and DecSccB = '1')) and
                       DevHit = '1' and SccModeB(4) = '0'
                  else '0';

  LSccModWe <= '1' when SccModWe = '1' and pSltAdr(9 downto 8) /= "11" else '0';
  RSccModWe <= '1' when SccModWe = '1' and pSltAdr(9 downto 8) /= "10" else '0';

  ----------------------------------------------------------------
  -- SPC / External memory (512kbytes * 2) access
  ----------------------------------------------------------------

  -- SCSI controller access enable (4000-5FFFh/A000-BFFFh)
  SccSpcCs <= '1' when pSltSltsl_n = '0' and pSltRst_n = '1' and SccModeA(6) = '1' and SccModeA(4) = '0' and
                       (pSltAdr(15 downto 12) = "0101" or pSltAdr(15 downto 12) = "1011") else '0';
  SccDmack <= '1' when pSltSltsl_n = '0' and pSltRst_n = '1' and SccModeA(6) = '1' and SccModeA(4) = '0' and
                       (pSltAdr(15 downto 12) = "0100" or pSltAdr(15 downto 12) = "1010") else '0';

  -- Memory access enable
  SccMemCs <= '1' when pSltSltsl_n = '0' and pSltRst_n = '1' and (
                       -- 4000-5FFFh(R)
                       (pSltAdr(15 downto 13) = "010" and SccModeA(6) = '0' and                   pSltRd_n = '0') or
                       -- 6000-7FFFh(R)
                       (pSltAdr(15 downto 13) = "011" and                                         pSltRd_n = '0') or
                       -- 8000-9FFFh(R)
                       (pSltAdr(15 downto 13) = "100" and                       DecSccA = '0' and pSltRd_n = '0') or
                       -- A000-BFFFh(R)
                       (pSltAdr(15 downto 13) = "101" and SccModeA(6) = '0' and DecSccB = '0' and pSltRd_n = '0') or
                       -- 4000-5FFFh(R/W) ESCC-RAM
                       (pSltAdr(15 downto 13) = "010" and SccModeA(4) = '1') or
                       -- 6000-7FFDh(R/W) ESCC-RAM
                       (pSltAdr(15 downto 13) = "011" and SccModeA(4) = '1' and Dec1FFE /= '1') or
                       -- 4000-7FFFh(R/W) SNATCHER
                       (pSltAdr(15 downto 14) = "01"  and SccModeB(4) = '1') or
                       -- 8000-9FFFh(R/W) SNATCHER
                       (pSltAdr(15 downto 13) = "100" and SccModeB(4) = '1') or
                       -- A000-BFFDh(R/W) SNATCHER
                       (pSltAdr(15 downto 13) = "101" and SccModeB(4) = '1' and Dec1FFE /= '1')
                  ) else '0';

  RamAdrEx  <= '1' when SccModeA(6) = '1' and pSltAdr(14 downto 13) /= "11" else '0';

  -- external port map

  pRamCeX_n <= '0' when SccMemCs = '1' and RamAdrEx = '0' else '1';
  pRamCeY_n <= '0' when SccMemCs = '1' and RamAdrEx = '1' else '1';

  pRamOeX_n <= '0' when pSltRd_n = '0' and pSltRst_n = '1' else '1';
  pRamOeY_n <= '0' when pSltRd_n = '0' and pSltRst_n = '1' else '1';
  pRamWeX_n <= '0' when pSltWr_n = '0' and pSltRst_n = '1' else '1';
  pRamWeY_n <= '0' when pSltWr_n = '0' and pSltRst_n = '1' else '1';

  pRamAdrX  <= SccBank0(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "10" else
               SccBank1(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "11" else
               SccBank2(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "00" else
               SccBank3(5 downto 0) & pSltAdr(12 downto 0);
  pRamAdrY  <= SccBank0(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "10" else
               SccBank1(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "11" else
               SccBank2(5 downto 0) & pSltAdr(12 downto 0) when pSltAdr(14 downto 13) = "00" else
               SccBank3(5 downto 0) & pSltAdr(12 downto 0);

  pRamDatX  <= pSltDat when pSltWr_n = '0' else (others => 'Z');
  pRamDatY  <= pSltDat when pSltWr_n = '0' else (others => 'Z');

  pSltDat   <= pRamDatX when SccMemCs = '1' and RamAdrEx = '0' and pSltRd_n = '0' else
               pRamDatY when SccMemCs = '1' and RamAdrEx = '1' and pSltRd_n = '0' else
               (others => 'Z');

  ----------------------------------------------------------------
  -- Connect components
  ----------------------------------------------------------------

  PsgCh  : psg_wave
    port map(
      pSltClk_n, pSltRst_n, PsgRegPtr, pSltDat, PsgAmp,
      PsgRegWe
    );

  SccCh  : scc_wave
    port map(
      pSltClk_n, pSltRst_n, pSltAdr(7 downto 0), pSltDat, SccAmp,
      SccRegWe, SccModWe, SccWavCe, SccWavOe, SccWavWe, SccWavWx, SccWavAdr, SccWavDat
    );

--@  SccChL : scc_wave
--    port map(
--      pSltClk_n, pSltRst_n, pSltAdr(7 downto 0), pSltDat, LSccAmp,
--      LSccRegWe, LSccModWe, LSccWavCe, LSccWavOe, LSccWavWe, LSccWavWx, SccWavAdr, SccWavDat
--    );

--@  SccChR : scc_wave
--    port map(
--      pSltClk_n, pSltRst_n, pSltAdr(7 downto 0), pSltDat, RSccAmp,
--      RSccRegWe, RSccModWe, RSccWavCe, RSccWavOe, RSccWavWe, RSccWavWx, SccWavAdr, SccWavDat
--    );

  ----------------------------------------------------------------
  -- 1 bit D/A  control
  ----------------------------------------------------------------
  process(pSltClk_n, pSltRst_n)

    variable AmpL : std_logic_vector(9 downto 0);
    variable AmpR : std_logic_vector(9 downto 0);
    variable Amp  : std_logic_vector(9 downto 0);
    variable AcuL : std_logic_vector(10 downto 0);
    variable AcuR : std_logic_vector(10 downto 0);
    variable Acu  : std_logic_vector(10 downto 0);

  begin

    if (pSltRst_n = '0') then

      AmpL := (others => '0');
      AmpR := (others => '0');
      Amp  := (others => '0');
      AcuL := (others => '0');
      AcuR := (others => '0');
      Acu  := (others => '0');
      pSltSndL  <= '0';
      pSltSndR  <= '0';
      pSltSound <= '0';

    elsif (pSltClk_n'event and pSltClk_n = '1') then

      AmpL := (('0' & PsgAmp & '0') + (LSccAmp & "00")) and "1111111000";
      AmpR := (('0' & PsgAmp & '0') + (RSccAmp & "00")) and "1111111000";
--@      Amp  := (('0' & PsgAmp & '0') + ('0' & LSccAmp & '0') + ('0' & RSccAmp & '0')) and "1111111000";
      Amp  := (('0' & PsgAmp & '0') + ( SccAmp & "00")) and "1111111000";
      AcuL := ('0' & AcuL(9 downto 0)) + ('0' & AmpL);
      AcuR := ('0' & AcuR(9 downto 0)) + ('0' & AmpR);
      Acu  := ('0' & Acu(9 downto 0)) + ('0' & Amp);
--@      pSltSndL <= AcuL(10) xor KeyClick xor '1';
--@      pSltSndR <= AcuR(10) xor KeyClick xor '1';
      pSltSndL  <= Acu(10) xor KeyClick xor '1';
      pSltSndR  <= Acu(10) xor KeyClick xor '1';
      pSltSound <= Acu(10) xor KeyClick xor '1';

    end if;

  end process;

end RTL;
